from typing import Optional, List, Type
from shreqt.model import ModelBase, ViewBase, full_table_name


class Layer:
    def __init__(
        self,
        *,
        schemas: Optional[List[str]] = None,
        tables: Optional[List[Type[ModelBase]]] = None,
        views: Optional[List[Type[ViewBase]]] = None,
        models: Optional[List[ModelBase]] = None,
    ):
        self.tables = tables or []
        self.schemas = schemas or []
        self.models = models or []
        self.views = views or []

    def to_list_push(self):
        return self.schemas + self.tables + self.views + self.models

    def to_list_pop(self):
        return self.models + self.views + self.tables + self.schemas


class LayerBuilder:
    def __init__(self):
        self.tables = {}
        self.schemas = set()
        self.models = []
        self.views = []

    def with_table(self, table: Type[ModelBase]):
        full_name = full_table_name(table)
        self.tables[full_name] = table
        self.with_schema(table.__table_args__["schema"])
        return self

    def with_schema(self, schema: str):
        self.schemas.add(schema)
        return self

    def with_view(self, view: Type[ViewBase]):
        self.views.append(view)
        return self

    def with_model(self, model: ModelBase):
        self.with_table(type(model))
        self.models.append(model)
        return self

    def build(self):
        return Layer(
            schemas=list(self.schemas),
            tables=list(self.tables.values()),
            models=self.models,
            views=self.views,
        )



import pyexasol
from contextlib import contextmanager
from os import environ


class ExasolConnection:
    @contextmanager
    def connect(self, schema=None):
        # revlibs.connections doesnt support schema = None.
        # Ideally we'd want to fix revlibs and use them instead of this.

        params = dict(compression=True)
        if schema:
            params["schema"] = schema

        connection = pyexasol.connect(
            dsn=environ.get("SHREQT_DSN", "localhost:8999"),
            user=environ.get("SHREQT_USER", "sys"),
            password=environ.get("SHREQT_PASS", "exasol"),
            fetch_dict=True,
            fetch_mapper=pyexasol.exasol_mapper,
            **params,
        )
        try:
            yield connection
        finally:
            connection.close()

    def ping(self):
        with self.connect() as conn:
            res = conn.execute("SELECT 1 FROM dual").fetchall()
            assert res[0]["1"] == 1



from shreqt.layer_query import layer_factory
from shreqt.layer import Layer
from enum import Enum
from sqlalchemy_exasol.pyodbc import EXADialect_pyodbc
from shreqt.connection import ExasolConnection
from typing import List


class DBOnion:
    class DBType(Enum):
        EXASOL = 1

    def __init__(self, db_type: DBType):
        self.db_type = db_type
        # FIXME, inject engine-based components once we support >1 db.
        self.connection = ExasolConnection()
        self._dialect = EXADialect_pyodbc()
        self._layers: List[Layer] = []

        # self.connection.ping() # Should not be in constructor

    def push_layer(self, layer: Layer):
        self._layers.append(layer)
        elements = layer.to_list_push()
        queries = [layer_factory(e, self._dialect).push_query(e) for e in elements]
        self._run_queries(queries)

    def pop_layer(self):
        layer = self._layers.pop()
        elements = layer.to_list_pop()
        queries = [layer_factory(e, self._dialect).pop_query(e) for e in elements]
        self._run_queries(queries)

    def _run_queries(self, queries: List[str]):
        with self.connection.connect() as conn:
            for query in queries:
                conn.execute(query)



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['shreqt']

package_data = \
{'': ['*']}

install_requires = \
['pyexasol>=0.6.4,<0.7.0',
 'sqlalchemy-exasol>=2.0,<3.0',
 'sqlalchemy>=1.3,<2.0',
 'sqlparse>=0.3.0,<0.4.0']

setup_kwargs = {
    'name': 'shreqt',
    'version': '0.2.2',
    'description': '',
    'long_description': None,
    'author': 'Zibi Rzepka',
    'author_email': 'zibi.rzepka@revolut.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql.expression import Executable, ClauseElement


class CreateView(Executable, ClauseElement):
    def __init__(self, name, select):
        self.name = name
        self.select = select


@compiles(CreateView)
def create_view(element, compiler, **kw):
    view_sql = compiler.process(element.select, literal_binds=True)
    return f"CREATE VIEW {element.name} AS \n{view_sql}"


class DropView(Executable, ClauseElement):
    def __init__(self, name):
        self.name = name


@compiles(DropView)
def drop_view(element, compiler, **kw):
    return f"DROP VIEW {element.name}"
